#include "GPAdvancedLayerAlgorithm.h"
#include "GPPhenotype.h"
#include "GPAdvancedLayerStruct.h"
#include <set>

using namespace std;

namespace genetic_pack {

	GPAdvancedLayerAlgorithm::GPAdvancedLayerAlgorithm(void)
	{
	}


	GPAdvancedLayerAlgorithm::~GPAdvancedLayerAlgorithm(void)
	{
	}

	// tries to put rectangle R into layer
	bool putIntoLayer(const GPContainer& container, const GPRectangle& R, 
		GPAdvancedLayerStruct& layer);

	// tries to put rectangle R into unallocated space
	bool putIntoUnallocatedSpace(const GPContainer& container, const GPRectangle& R,
		vector<GPAdvancedLayerStruct>& layers);

	// AdvancedLayer algorithm implementation
	// decoding cromosome into phenotype
	// i.e. generating layer packing from permutation
	GPPhenotype GPAdvancedLayerAlgorithm::decodeChromosome(const GPContainer& container, 
		const GPChromosome& chromo)
	{
		vector<GPAdvancedLayerStruct> layers;
		for (unsigned int i = 0; i < chromo.size(); i++)
		{
			GPRectangle R = container.GetRectangles()[chromo[i]];
			// trying to put the rectangle into layers
			bool foundLayer = false;
			for (vector<GPAdvancedLayerStruct>::iterator layer = layers.begin(); !foundLayer && layer != layers.end(); layer++)
			{
				if (putIntoLayer(container, R, *layer))
				{
					foundLayer = true;
					break;
				}
			} // for layer
			if (!foundLayer) {
				putIntoUnallocatedSpace(container, R, layers);
			}
		}

		// converting Layers into phenotype
		// i.e. putting all the rectangles from all the layers together
		vector<GPRectPosition> rectangles;
		for (vector<GPAdvancedLayerStruct>::const_iterator layer = layers.begin(); layer != layers.end(); layer++)
			for (vector<GPRectPosition>::const_iterator rect = layer->getRectangles().begin(); 
				rect != layer->getRectangles().end(); rect++)
			{
				rectangles.push_back(*rect);
			}

		GPPhenotype result(container, rectangles);
		return result;
	}

	// tries to put rectangle R into layer
	bool putIntoLayer(const GPContainer& container, const GPRectangle& R, 
		GPAdvancedLayerStruct& layer)
	{
		// for faster checking
		if (R.GetHeight() > container.GetMetalSheet().GetHeight() || R.GetWidth() > layer.getRight() - layer.getLeft() ||
			layer.getCoveredArea() + R.GetArea() > (layer.getRight() - layer.getLeft()) * layer.getHeight())
		{
			return false;
		}

		bool placeFound = false;
		for (set<double>::iterator itY = layer.getY().begin(); 
			!placeFound && itY != layer.getY().end(); ++itY)
		{
			for (set<double>::iterator itX = layer.getX().begin(); 
				!placeFound && itX != layer.getX().end(); ++itX)
			{
				GPPoint base;
				base.x = *itX;
				base.y = *itY;
				GPRectPosition place(R, base);
				if (layer.tryToPut(place))
				{
					placeFound = true;
				}
			}
		}
		
		return placeFound;
	}

	// tries to put rectangle R into unallocated space
	bool putIntoUnallocatedSpace(const GPContainer& container, const GPRectangle& R,
		vector<GPAdvancedLayerStruct>& layers)
	{
		GPPoint usBase;
		usBase.x = layers.size() > 0 ? layers[layers.size()-1].getRight() : 0.0;
		usBase.y = 0;

		// if is possible to create one more layer
		if (compare(usBase.x, container.GetMetalSheet().GetWidth()) >= 0)
		{
			return false;
		}

		// if it is enough free area to put R
		if (R.GetArea() > (container.GetMetalSheet().GetWidth() - usBase.x) * container.GetMetalSheet().GetHeight())
		{
			return false;
		}

		GPRectPosition usRect(GPRectangle(container.GetMetalSheet().GetWidth() - usBase.x, 
			container.GetMetalSheet().GetHeight()), usBase);
		vector<GPPoint> defects;

		// selecting defects which are in the unallocated space
		for (vector<GPPoint>::const_iterator itD = container.GetDefects().begin(); 
			itD != container.GetDefects().end(); ++itD)
		{
			if (isStrictlyInsideRect(*itD, usRect))
			{
				defects.push_back(*itD);
			}
		}

		// creating layer representing unallocated space
		GPAdvancedLayerStruct usLayer(container.GetMetalSheet().GetWidth(), container.GetMetalSheet().GetHeight(),
			usBase.x, container.GetMetalSheet().GetWidth(), defects);

		// trying to put into unallocated space layer
		bool placeFound = false;
		for (set<double>::iterator itX = usLayer.getX().begin(); 
			!placeFound && itX != usLayer.getX().end(); ++itX)
		{
			for (set<double>::iterator itY = usLayer.getY().begin(); 
				!placeFound && itY != usLayer.getY().end(); ++itY)
			{
				GPPoint base;
				base.x = *itX;
				base.y = *itY;
				GPRectPosition place(R, base);
				if (usLayer.tryToPut(place))
				{
					placeFound = true;
				}
			}
		}		

		if (!placeFound)
		{
			return false;
		}

		// constructing minimal layer required for unallocated space
		GPRectPosition usMinRect(GPRectangle(usLayer.getRectangles()[0].GetTopRight().x - usBase.x, 
			container.GetMetalSheet().GetHeight()), usBase);
		vector<GPPoint> minDefects;

		for (vector<GPPoint>::iterator itD = defects.begin(); itD != defects.end(); ++itD)
		{
			if (isStrictlyInsideRect(*itD, usMinRect))
			{
				minDefects.push_back(*itD);
			}
		}

		GPAdvancedLayerStruct usMinLayer(container.GetMetalSheet().GetWidth(), container.GetMetalSheet().GetHeight(),
			usBase.x, usMinRect.GetTopRight().x, minDefects);

		if (!usMinLayer.tryToPut(usLayer.getRectangles()[0]))
		{
			cerr << "Redundant error check on putting rectangle in an unallocated space";
			throw "Redundant error check on putting rectangle in an unallocated space";
		}

		layers.push_back(usMinLayer);

		return true;
	}

} // namespace genetic_pack
