#include "GPAlgorithm.h"
#include "GPRandom.h"
#include "GPDiscreteRandom.h"

#include <sstream>
#include <vector>
#include <cmath>

using namespace std;

namespace genetic_pack {

	GPAlgorithm::GPAlgorithm(const GPContainer& container, unsigned int popSize, unsigned int eliteSize, 
		double mutationProp,  int crossoverMode, int layoutMethod):
	container(container), popSize(popSize), eliteSize(eliteSize), mutationProp(mutationProp), 
		crossoverMode(crossoverMode), layoutMethod(layoutMethod)
	{
		ostringstream s;
		if (popSize < 2)
		{
			s << "GPAlgorithm error: incorrect size of population (" << popSize << ")";
			throw s.str();
		}
		if (eliteSize < 2 || eliteSize >= popSize)
		{
			s << "GPAlgorithm error: incorrect size of reproducing (" << eliteSize << ")";
			throw s.str();
		}
		if (mutationProp < 0 || mutationProp > 1)
		{
			s << "GPAlgorithm error: incorrect mutation probability (" << mutationProp << ")";
			throw s.str();
		}
		int n = container.GetRectangles().size();
		if (log(popSize) > n*(log(n) - 1) +1)
		{
			s << "GPAlgorithm error: population size (" << popSize << ") is larger than number of possible chromosomes";
			throw s.str();
		}

		phenotypesCached = false;
	}

	void GPAlgorithm::initPopulation()
	{
		phenotypesCached = false;

		population.clear();

		// TODO: optimize that

		// generating random chromosomes, i.e. permutations
		for (unsigned int i = 0; i < popSize; i++)
		{
			int sz = container.GetRectangles().size();
			GPChromosome randChromo = randomChromosome(sz);
			// ensuring the chromosome has not been added still
			while (randChromo.find(population) > -1)
				randChromo = randomChromosome(sz);
			population.push_back(randChromo);
		}
	}

	void GPAlgorithm::nextGeneration()
	{
		if (!phenotypesCached)
			cachePhenotypes();

		// sort chromosomes and phenotypes by fitness (so that population[0] has the largest)
		for (unsigned int i = 0; i < population.size(); i++)
		{
			for (unsigned int j = i+1; j < population.size(); j++)
			{
				if (phenotypes[i].fitnessFunction() < phenotypes[j].fitnessFunction())
				{
					GPChromosome tmpChromo = population[i];
					population[i] = population[j];
					population[j] = tmpChromo;
					GPPhenotype tmpPheno = phenotypes[i];
					phenotypes[i] = phenotypes[j];
					phenotypes[j] = tmpPheno;
				};
			}
		}

		vector<GPChromosome> newPopulation;

		// reproducing eliteSize chromosomes
		for (unsigned int i = 0; i < eliteSize; i++)
			newPopulation.push_back(population[i]);

		vector<double> fits;
		for (unsigned int i = 0; i < eliteSize; i++)
			fits.push_back(phenotypes[i].fitnessFunction());

		GPDiscreteRandom weightedRnd(fits);

		// add new chromosomes produced by crossover
		for (unsigned int i = eliteSize; i < popSize; i++)
		{
			GPChromosome child = population[0]; // meaningless, just to create variable
			do
			{
				// father selection is weighted according to the fitness function
				GPChromosome father = newPopulation[weightedRnd.GetDiscrete()];

				// mother selection is just random from previous generation
				GPChromosome mother = newPopulation[getRandom(0, eliteSize)];
				while (mother == father)
					mother = newPopulation[getRandom(0, eliteSize)];

				// crossover
				child = doCrossover(father, mother, crossoverMode);

				// possibly mutation
				if (getRandom() < mutationProp)
					child.doMutation();
			}
			while (child.find(newPopulation) > -1); // ensures the child is not in the new population yet

			newPopulation.push_back(child);
		}

		population = newPopulation;

		phenotypesCached = false;
	}

	void GPAlgorithm::doAllSteps(int generations)
	{
#ifdef _DEBUG
		cout << "DEBUG:init population\n";
#endif
		initPopulation();
		for (int i = 0; i < generations; i++)
		{
#ifdef _DEBUG
			cout << "DEBUG: next generation\n";
#endif
			nextGeneration();
		}
	}

	double GPAlgorithm::GetBestFit()
	{
		double fit = -1.0;

		if (!phenotypesCached)
			cachePhenotypes();

		for (vector<GPPhenotype>::const_iterator it  = phenotypes.begin(); it != phenotypes.end(); it++)
			if (fit < it->fitnessFunction())
				fit = it->fitnessFunction();

		return fit;
	}

	double GPAlgorithm::GetAverageFit()
	{
		double sumFit = 0.0;

		if (!phenotypesCached)
			cachePhenotypes();

		for (vector<GPPhenotype>::const_iterator it = phenotypes.begin(); it != phenotypes.end(); it++)
			sumFit += it->fitnessFunction();

		return sumFit / phenotypes.size();
	}

	GPChromosome GPAlgorithm::GetBestChromosome()
	{
		if (!phenotypesCached)
			cachePhenotypes();

		double fit = phenotypes[0].fitnessFunction();
		GPChromosome best = population[0];

		for (unsigned int i = 1; i < phenotypes.size(); i++)
		{
			if (fit < phenotypes[i].fitnessFunction())
			{
				fit = phenotypes[i].fitnessFunction();
				best = population[i];
			}
		}

		return best;
	}

	// optimisation 
	void GPAlgorithm::cachePhenotypes()
	{
		phenotypes.clear();

		for (vector<GPChromosome>::const_iterator it = population.begin(); it != population.end(); it++)
			phenotypes.push_back(decodeChromosome(container, *it, layoutMethod));

		phenotypesCached = true;
	}
} // namespace genetic_pack