#include "GPChromosome.h"
#include "GPRandom.h"

#include <sstream>
#include <set>
#include <queue>
#include <iostream>

using namespace std;

namespace genetic_pack {

	GPChromosome::GPChromosome(const vector<unsigned int> permutation): permutation(permutation)
	{
		// checking that permutation indeed represents a permutation
		// of (0,1,2,3,...,n-1)
		vector<bool> presents;
		for (unsigned int i = 0; i < permutation.size(); i++)
		{
			presents.push_back(false);
		}

		for (unsigned int i = 0; i < permutation.size(); i++)
		{
			if (permutation[i] < 0 || permutation[i] >= permutation.size() || presents[permutation[i]])
			{
				throw "GPChromosome error: permutation is incorrect";
			}
			presents[permutation[i]] = true;
		}
	}

	// simple mutation: swaps two random positions in permutation
	void GPChromosome::doMutation()
	{
		unsigned int i = getRandom(0, permutation.size());
		unsigned int j = getRandom(0, permutation.size());
		while (j == i)
			j = getRandom(0, permutation.size());
		unsigned int tmp = permutation[i];
		permutation[i] = permutation[j];
		permutation[j] = tmp;
	}

	// generating random chromosome, i.e. permutation
	GPChromosome randomChromosome(unsigned int n)
	{
		vector<unsigned int> permutation;
		for (unsigned int i = 0; i < n; i++)
			permutation.push_back(i);

		for (unsigned int i = 0; i < n-1; i++)
		{
			int j = getRandom(i+1, n);
			unsigned int tmp = permutation[i];
			permutation[i] = permutation[j];
			permutation[j] = tmp;
		}

		return GPChromosome(permutation);
	}

	GPChromosome doCrossover(const GPChromosome& a, const GPChromosome& b, int mode)
	{
		if (a.GetPermutation().size() != b.GetPermutation().size())
		{
			ostringstream s;
			s << "GPChromosome.doCrossover() error: chromosomes have different sizes";
			throw s.str();
		}

		switch (mode)
		{
		case JAKOBS_CROSSOVER: return doJakobsCrossover(a, b);
		case PMX_CROSSOVER: return doPMXCrossover(a, b);
		default:
			ostringstream s;
			s << "GPChromosome.doCrossover() error: mode " << mode << " is either undefined or not implemented still";
			throw s.str();
		}
	}

	GPChromosome doJakobsCrossover(const GPChromosome& a, const GPChromosome& b)
	{		
		int n = a.GetPermutation().size();
		int p = getRandom(0, n), q = getRandom(0, n);

		vector<unsigned int> newPermutation;
		set<int> pushed;
		
		for (int i = p; i < p+q; i++)
		{
			newPermutation.push_back(a.GetPermutation()[i % n]);
			pushed.insert(a.GetPermutation()[i % n]);
		}

		for (int i = 0; i < n; i++)
		{
			if (pushed.count(b.GetPermutation()[i]) == 0)
			{
				newPermutation.push_back(b.GetPermutation()[i]);
			}
		}

		GPChromosome chromosome(newPermutation);

		return chromosome;
	}

	// particular crossover - PMX
	GPChromosome doPMXCrossover(const GPChromosome& a, const GPChromosome& b)
	{
		int n = a.GetPermutation().size();
		int first = getRandom(0, n), last = getRandom(first, n);

		vector<unsigned int> newPermutation;
		// initialising vector, only to have required size of it
		for (int i = 0; i < n; i++)
			newPermutation.push_back(0xFF);

		queue<unsigned int> matchQueue;
		set<unsigned int> matchSet;

		for (int i = first; i < last; i++)
		{
			matchSet.insert(a.GetPermutation()[i]);
		}

		for (int i = first; i < last; i++)
		{
			if (matchSet.count(b.GetPermutation()[i]) == 0)
				matchQueue.push(b.GetPermutation()[i]);
			newPermutation[i] = a.GetPermutation()[i];
		}

		for (int i = 0; i < n; i++)
		{
			if (i < first || i >= last)
			{
				if (matchSet.count(b.GetPermutation()[i]) > 0)
				{
					newPermutation[i] = matchQueue.front();
					matchQueue.pop();
				}
				else
				{
					newPermutation[i] = b.GetPermutation()[i];
				}
			}
		}

		GPChromosome res(newPermutation);
		
		return res;
	}

	// convinience operator [], indexing permutation elements
	unsigned int GPChromosome::operator[](unsigned int index) const
	{
		if (index < 0 || index >= permutation.size())
		{
			ostringstream s;
			s << "GPChromosome[] error: index(" << index << ") is out of bounds";
			throw s.str();
		}

		return permutation[index];
	}

	bool GPChromosome::operator==(const GPChromosome& another) const
	{
		if (permutation.size() != another.GetPermutation().size())
			return false;

		for (unsigned int i = 0; i < permutation.size(); i++)
			if (permutation[i] != another.GetPermutation()[i])
				return false;
		return true;
	}

	int GPChromosome::find(const vector<GPChromosome>& lst)
	{
		for (unsigned int i = 0; i < lst.size(); i++)
			if (*this == lst[i])
				return i;
		return -1;
	}

} // namespace genetic_pack