#include "GPParams.h"
#include "GPChromosome.h"
#include "GPLayoutMethod.h"

#include <algorithm>
#include <string>
#include <sstream>
#include <iostream>

namespace genetic_pack {

	GPParams::GPParams(int argc, char* argv[]):
		crossoverMode(JAKOBS_CROSSOVER),
		layoutMethod(SIMPLE_LAYER),
		mutationProb(0.1),
		popSize(20),
		eliteSize(5),
		generations(100),
		rotateAndRepeat(false),
		tries(1),
		rotateSheet(false)
	{
		// converting params to key=value map
		pm = paramsMap(argc, argv);

		// reading params from the map
		if (pm.count("crossovermode") > 0)
		{
			if (pm.find("crossovermode")->second == "jakobs")
				crossoverMode = JAKOBS_CROSSOVER;
			if (pm.find("crossovermode")->second == "pmx")
				crossoverMode = PMX_CROSSOVER;
		}

		if (pm.count("layoutmethod") > 0)
		{
			if (pm.find("layoutmethod")->second == "simple-layer")
				layoutMethod = SIMPLE_LAYER;
			if (pm.find("layoutmethod")->second == "advanced-layer")
				layoutMethod = ADVANCED_LAYER;
		}

		if (pm.count("ifname") > 0)
		{
			ifname = pm.find("ifname")->second;
		}
		else
		{
			cerr << "Incorrect usage. At least ifname should be defined" << endl;
			throw "Incorrect usage. At least ifname should be defined";
		}

		parseDouble("mutationprob", mutationProb);
		parseInt("popsize", popSize);
		parseInt("elitesize", eliteSize);
		parseInt("generations", generations);
		parseBool("rotateandrepeat", rotateAndRepeat);
		parseInt("tries", tries);
		parseBool("rotatesheet", rotateSheet);

		if (eliteSize >= popSize)
		{
			cerr << "Elite size cannot be more than population size\n";
			throw "Elite size cannot be more than population size";
		}
	}

	void GPParams::dumpParams(ostream& out) const {
		out << "BEGIN DUMP PARAMETERS\n";
		
		switch (crossoverMode)
		{
		case JAKOBS_CROSSOVER:
			out << "crossoverMode = jakobs\n";
			break;
		case PMX_CROSSOVER:
			out << "crossoverMode = pmx\n";
		default:
			out << "crossoverMode = #ERROR\n";
			break;
		}

		switch (layoutMethod)
		{
		case SIMPLE_LAYER:
			out << "layoutMethod = simple-layer\n";
			break;
		case ADVANCED_LAYER:
			out << "layoutMethod = advanced-layer\n";
			break;
		default:
			out << "layoutMethod = #ERROR\n";
			break;
		}

		out << "mutationProb = " << mutationProb << '\n';
		out << "popSize = " << popSize << '\n';
		out << "eliteSize = " << eliteSize << '\n';
		out << "generations = " << generations << '\n';
		out << "rotateAndRepeat = " << rotateAndRepeat << '\n';
		out << "tries = " << tries  << '\n';
		out << "rotateSheet = " << rotateSheet << '\n';
		out << "END DUMP PARAMETERS" << endl;
	}

	map<string, string> GPParams::paramsMap(int argc, char* argv[])
	{
		map<string, string> pm;

		for (int i = 1; i < argc; i++)
		{
			string s(argv[i]);
			int delimPos = s.find_first_of('=');
			if (delimPos == string::npos)
				continue;
			string key = s.substr(0, delimPos);
			string value = s.substr(delimPos+1, s.length() - delimPos - 1);
			if (s.length() > 1 && s[0] == '"' && s[s.length()-1] == '"')
				value = s.substr(1, s.length()-2);

			// converting both key and value to lower case in order to ignore case of params
			std::transform(key.begin(), key.end(), key.begin(), ::tolower);
			std::transform(value.begin(), value.end(), value.begin(), ::tolower);

			pm[key] = value;
		}

		return pm;
	}

	void GPParams::parseString(string key, string& value) {
		if (pm.count(key) > 0)
			value = pm.find(key)->second;
	}

	void GPParams::parseInt(string key, int& value) {
		if (pm.count(key) > 0)
		{
			stringstream tmp(pm.find(key)->second);
			tmp >> value;
		}
	}

	void GPParams::parseDouble(string key, double& value) {
		if (pm.count(key) > 0)
		{
			stringstream tmp(pm.find(key)->second);
			tmp >> value;
		}
	}

	void GPParams::parseBool(string key, bool& value) {
		if (pm.count(key) > 0)
			value = pm.find(key)->second == "true";
	}
}