#include "GPSimpleLayerAlgorithm.h"

#include <set>

namespace genetic_pack {

	GPSimpleLayerAlgorithm::GPSimpleLayerAlgorithm(void)
	{
	}


	GPSimpleLayerAlgorithm::~GPSimpleLayerAlgorithm(void)
	{
	}

	// internal structure to work with layer
	struct Layer
	{
		double left, right, height;
		vector<GPRectPosition> rectangles;
	};

	// tries to put rectangle R into layer
	bool TryToPut(const GPContainer& container, const GPRectangle& R, Layer& layer)
	{
		if (layer.right - layer.left < R.GetWidth())
			return false;

		// trying to put R into the layer
		double bottom = layer.height;
		double top = container.GetMetalSheet().GetHeight();

		// while R is not shifted too high
		while (R.GetHeight() <= top - bottom)
		{
			vector<GPPoint> D; // defects covered by R
			for (vector<GPPoint>::const_iterator defect = container.GetDefects().begin(); 
				defect != container.GetDefects().end(); defect ++)
			{
				// if defect is covered by R, add it to D
				if (layer.left < defect->x && defect->x < layer.right &&
					bottom < defect->y && defect->y < bottom + R.GetHeight())
				{
					D.push_back(*defect);
				}
			}

			set<double> x; // set of x'es of defects in D
			x.insert(layer.left);
			x.insert(layer.right);
			for (unsigned int i = 0; i < D.size(); i++)
			{
				x.insert(D[i].x);
			}

			bool found = false;
			double x1, x2;
			// trying to find x1 and x2 such that R could be put between them
			for (set<double>::const_iterator it = x.begin(); it != x.end();)
			{
				x1 = *it;
				++it;
				if (it == x.end())
					break;
				x2 = *it;
				if (x2 - x1 >= R.GetWidth())
				{
					found = true;
					break;
				}
			}

			if (!found)
			{
				// no such x1 and x2 found, shifting up so that R is at minimum of y'es from D
				double newBottom = D[0].y;
				for (unsigned int i = 1; i < D.size(); i++)
					newBottom = (newBottom > D[i].y) ? D[i].y : newBottom;
				bottom = newBottom;
			}
			else
			{
				// Place for R found
				layer.height = bottom + R.GetHeight();
				GPPoint position = {x1, bottom};
				layer.rectangles.push_back(GPRectPosition(R, position));
				return true;
			}
		}
		return false;
	}

	// tries to put R into unallocated space
	bool TryToPut(const GPContainer& container, vector<Layer>& Layers, const GPRectangle& R)
	{
		if (R.GetHeight() > container.GetMetalSheet().GetHeight()) // redundant check -> just in case
			return false;

		// trying to put R into unallocated space
		double left = (Layers.size() > 0) ? (Layers[Layers.size()-1].right) : 0;
		double right = container.GetMetalSheet().GetWidth();

		// while R is not shifted right too much
		while (R.GetWidth() <= right-left)
		{
			vector<GPPoint> D; // defects on potential layer
			for (vector<GPPoint>::const_iterator defect = container.GetDefects().begin(); 
				defect != container.GetDefects().end(); defect ++)
				if (left < defect->x && defect->x < left + R.GetWidth())
					D.push_back(*defect);

			set<double> y; // y'es of defects from D
			y.insert(0);
			y.insert(container.GetMetalSheet().GetHeight());
			for (unsigned int i = 0; i < D.size(); i++)
				y.insert(D[i].y);

			bool found = false;
			double y1, y2;
			// looking for y1 and y2 such that R could lie between them
			for (set<double>::const_iterator it = y.begin(); it != y.end();)
			{
				y1 = *it;
				++it;
				if (it == y.end())
					break;
				y2 = *it;
				if (y2 - y1 >= R.GetHeight())
				{
					found = true;
					break;
				}
			}

			if (!found)
			{
				// not founf, shifting to the minimum x from D
				double newLeft = D[0].x;
				for (unsigned int i = 1; i < D.size(); i++)
					newLeft = (newLeft > D[i].x) ? D[i].x : newLeft;
				left = newLeft;
			}
			else
			{
				// potential layer found, creating it
				Layer newLayer;
				newLayer.left = (Layers.size() > 0) ? (Layers[Layers.size()-1].right) : 0;
				newLayer.right = left + R.GetWidth();
				newLayer.height = y1 + R.GetHeight();
				GPPoint position = {left, y1};
				newLayer.rectangles.push_back(GPRectPosition(R, position));
				Layers.push_back(newLayer);
				return true;
			}
		}
		return false;
	}

	// SimpleLayer algorithm implementation
	// decoding cromosome into phenotype
	// i.e. generating layer packing from permutation
	GPPhenotype GPSimpleLayerAlgorithm::decodeChromosome(const GPContainer& container, 
		const GPChromosome& chromo)
	{
		vector<Layer> Layers;
		for (unsigned int i = 0; i < chromo.size(); i++)
		{
			GPRectangle R = container.GetRectangles()[chromo[i]];
				// trying to put the rectangle into layers
			bool foundLayer = false;
			for (vector<Layer>::iterator layer = Layers.begin(); !foundLayer && layer != Layers.end(); layer++)
			{
				if (TryToPut(container, R, *layer))
				{
					foundLayer = true;
					break;
				}
			} // for layer
			if (!foundLayer)
				TryToPut(container, Layers, R);
		}

		// converting Layers into phenotype
		// i.e. putting all the rectangles from all the layers together
		vector<GPRectPosition> rectangles;
		for (vector<Layer>::const_iterator layer = Layers.begin(); layer != Layers.end(); layer++)
			for (vector<GPRectPosition>::const_iterator rect = layer->rectangles.begin(); rect != layer->rectangles.end(); rect++)
				rectangles.push_back(*rect);

		GPPhenotype result(container, rectangles);
		return result;
	}

} // namespace genetic_pack