#include <iostream>
#include <fstream>
#include <map>
#include <sstream>
#include <algorithm>
#include <ctime>

#include "GPAlgorithm.h"
#include "GPPhenotype.h"
#include "GPRectangle.h"
#include "GPPoint.h"
#include "GPParams.h"

using namespace std;
using namespace genetic_pack;

GPContainer readContainer(istream &in, bool rotateAndRepeat);
GPContainer rotateContainer(GPContainer baseContainer, int mode);
GPPoint rotatePoint(GPPoint p, double W, double H, int mode);


// the main function of console program
int main(int argc, char* argv[])
{
	// start time
	time_t startTime = time(NULL);

	GPParams consoleParams(argc, argv);

	ifstream fin(consoleParams.ifname);
	GPContainer baseContainer = readContainer(fin, consoleParams.rotateAndRepeat);
	fin.close();

	ofstream intermediateResults("intermediate-results.txt");
	consoleParams.dumpParams(intermediateResults);
	consoleParams.dumpParams(cout);

	double globalBestFit = -1;

	for (int mode = 0; mode < (consoleParams.rotateSheet ? 8 : 1); mode++)
	{
		intermediateResults << "Metal sheet rotation mode " << mode << endl;
		cout << "Metal sheet rotation mode " << mode << endl;
		GPContainer container = rotateContainer(baseContainer, mode);
		for (int tryNo = 0; tryNo < consoleParams.tries; tryNo++)
		{
			GPAlgorithm alg(container, consoleParams.popSize, consoleParams.eliteSize,
				consoleParams.mutationProb, consoleParams.crossoverMode, consoleParams.layoutMethod);
			alg.doAllSteps(consoleParams.generations);
			intermediateResults << '\t' << "try #" << tryNo << ": "
				<< "best=" << alg.GetBestFit() 
				<< " (" << ((int)(alg.GetBestFit() / container.GetMetalSheet().GetHeight() 
				/ container.GetMetalSheet().GetWidth() * 100)) << "%)" << endl;
			cout << '\t' << "try #" << tryNo << ": "
				<< "best=" << alg.GetBestFit() 
				<< " (" << ((int)(alg.GetBestFit() / container.GetMetalSheet().GetHeight() 
				/ container.GetMetalSheet().GetWidth() * 100)) << "%)" << endl;
			if (alg.GetBestFit() > globalBestFit)
			{
				// output new best layout
				GPPhenotype layout = decodeChromosome(container, alg.GetBestChromosome(), 
					consoleParams.layoutMethod);
				
				ofstream fout;
				fout.open("sheet_metal.txt");
				fout << container.GetMetalSheet().GetWidth() << ' ' << container.GetMetalSheet().GetHeight() << endl;
				fout.close();

				fout.open("rectangles.txt");
				for (unsigned int i = 0; i < layout.GetRectangles().size(); i++)
				{
					GPRectPosition R = layout.GetRectangles()[i];
					fout << R.GetRectangle().GetWidth() << " " << R.GetRectangle().GetHeight() << " " << 
						R.GetBase().x << " " << R.GetBase().y << "\n";
				}
				fout.close();

				fout.open("defects.txt");
				for (unsigned int i = 0; i < container.GetDefects().size(); i++)
				{
					fout << container.GetDefects()[i].x << ' ' << container.GetDefects()[i].y << endl;
				}
				fout.close();

				fout.open("permutation.txt");
				for (unsigned int i = 0; i < alg.GetBestChromosome().GetPermutation().size(); i++)
				{
					fout << alg.GetBestChromosome().GetPermutation()[i] << ' ';
				}
				fout.close();


				globalBestFit = alg.GetBestFit();
			}
		}
	}

	intermediateResults << "BEST RESULT: " << globalBestFit << " (" << 
		((int)(globalBestFit / baseContainer.GetMetalSheet().GetHeight() / 
		baseContainer.GetMetalSheet().GetWidth() * 100)) << "%)" << endl;
	cout << "BEST RESULT: " << globalBestFit << " (" << 
		((int)(globalBestFit / baseContainer.GetMetalSheet().GetHeight() / 
		baseContainer.GetMetalSheet().GetWidth() * 100)) << "%)" << endl;
	intermediateResults.close();

	return 0;
}

// reads container data from the data
GPContainer readContainer(istream &in, bool rotateAndRepeat)
{
	double W, H;
	in >> W >> H;
	GPRectangle metalSheet(W, H);

	int numRect;
	in >> numRect;
	vector<GPRectangle> rectangles;
	for (int i = 0; i < numRect; i++)
	{
		double w, h;
		in >> w >> h;
		int repetitionNum = rotateAndRepeat ? ((int)(W * H / w / h)) : 1;
		// repeat rectangles (if specified)
		for (int repeat = 0; repeat < repetitionNum; repeat++)
		{
			rectangles.push_back(GPRectangle(w, h));
			if (rotateAndRepeat) // adding rotated copy
				rectangles.push_back(GPRectangle(h, w));
		}
	}

	vector<GPPoint> defects;
	GPPoint p;
	int numDefects;
	in >> numDefects;
	for (int i = 0; i < numDefects; i++)
	{
		in >> p.x >> p.y;
		defects.push_back(p);
	}

	GPContainer container(metalSheet, defects, rectangles);

	return container;
}

// rotates the container into required mode
GPContainer rotateContainer(GPContainer baseContainer, int mode)
{
	// metal sheet rotation
	double W, H;
	switch (mode)
	{
	case 2: case 3: case 6: case 7: 
		W = baseContainer.GetMetalSheet().GetHeight(); 
		H = baseContainer.GetMetalSheet().GetWidth();
		break;
	default:
		H = baseContainer.GetMetalSheet().GetHeight(); 
		W = baseContainer.GetMetalSheet().GetWidth();
		break;
	}
	GPRectangle metalSheet(W, H);

	// defects rotation
	vector<GPPoint> defects;
	for (unsigned int i = 0; i < baseContainer.GetDefects().size(); i++)
	{
		defects.push_back(rotatePoint(baseContainer.GetDefects()[i], 
			baseContainer.GetMetalSheet().GetWidth(), baseContainer.GetMetalSheet().GetHeight(), 
			mode));
	}

	// rectangles rotation
	vector<GPRectangle> rectangles;
	switch (mode)
	{
	case 2: case 3: case 6: case 7:
		for (unsigned int i = 0; i < baseContainer.GetRectangles().size(); i++)
		{
			rectangles.push_back(GPRectangle(baseContainer.GetRectangles()[i].GetHeight(),
				baseContainer.GetRectangles()[i].GetWidth()));
		}
		break;
	default:
		rectangles = baseContainer.GetRectangles();
		break;
	}

	return GPContainer(metalSheet, defects, rectangles);
}

// rotates the point into required mode
GPPoint rotatePoint(GPPoint p, double W, double H, int mode)
{
	double x, y;
	switch (mode)
	{
	case 1:
		x = W - p.x; y = p.y;
		break;
	case 2:
		x = p.y; y = p.x;
		break;
	case 3:
		x = H - p.y; y = p.x;
		break;
	case 4:
		x = p.x; y = H - p.y;
		break;
	case 5:
		x = W - p.x; y = H - p.y;
		break;
	case 6:
		x = p.y; y = W - p.x;
		break;
	case 7:
		x = H - p.y; y = W - p.x;
		break;
	default:
		x = p.x; y = p.y;
		break;
	}

	GPPoint newP;
	newP.x = x;
	newP.y = y;

	return newP;
}